@@ -36,9 +36,22 @@ inline constexpr const int32_t kIOVMax = 1024;
 inline constexpr const int32_t kMaxSaDataLen = 26;
 } // namespace detail
 
+struct WasiAddrStorage {
+  uint16_t AddressFamily;
+  uint8_t Address[128 - sizeof(uint16_t)];
+  __wasi_address_family_t getAddressFamily() const noexcept {
+    return static_cast<__wasi_address_family_t>(AddressFamily);
+  }
+  void setAddressFamily(__wasi_address_family_t AddrFamily) noexcept {
+    AddressFamily = static_cast<uint16_t>(AddrFamily);
+  }
+  Span<uint8_t> getAddress() noexcept { return Address; }
+  Span<const uint8_t> getAddress() const noexcept { return Address; }
+};
+static_assert(sizeof(WasiAddrStorage) == 128, "wrong size");
+
 class EVPoller;
-class EVEpoller;
-class Environ {
+class Environ : public PollerContext {
 public:
   ~Environ() noexcept;
 
@@ -88,14 +101,17 @@ public:
   WasiExpect<void> argsGet(Span<uint8_t_ptr> Argv,
                            Span<uint8_t> ArgvBuffer) const noexcept {
     for (const auto &Argument : Arguments) {
-      const __wasi_size_t Size =
-          static_cast<__wasi_size_t>(Argument.size()) + UINT32_C(1);
+      const __wasi_size_t Size = static_cast<__wasi_size_t>(Argument.size());
       std::copy_n(Argument.begin(), Size, ArgvBuffer.begin());
-      ArgvBuffer = ArgvBuffer.subspan(Size);
-      Argv[1] = Argv[0] + Size;
+      ArgvBuffer[Size] = '\0';
+      ArgvBuffer = ArgvBuffer.subspan(Size + UINT32_C(1));
+      if (Argv.size() > 1) {
+        Argv[1] = Argv[0] + Size + UINT32_C(1);
+      }
       Argv = Argv.subspan(1);
     }
-    Argv[0] = 0;
+    assert(ArgvBuffer.empty());
+    assert(Argv.empty());
 
     return {};
   }
@@ -131,13 +147,17 @@ public:
                               Span<uint8_t> EnvBuffer) const noexcept {
     for (const auto &EnvironVariable : EnvironVariables) {
       const __wasi_size_t Size =
-          static_cast<__wasi_size_t>(EnvironVariable.size()) + UINT32_C(1);
+          static_cast<__wasi_size_t>(EnvironVariable.size());
       std::copy_n(EnvironVariable.begin(), Size, EnvBuffer.begin());
-      EnvBuffer = EnvBuffer.subspan(Size);
-      Env[1] = Env[0] + Size;
+      EnvBuffer[Size] = '\0';
+      EnvBuffer = EnvBuffer.subspan(Size + UINT32_C(1));
+      if (Env.size() > 1) {
+        Env[1] = Env[0] + Size + UINT32_C(1);
+      }
       Env = Env.subspan(1);
     }
-    Env[0] = 0;
+    assert(EnvBuffer.empty());
+    assert(Env.empty());
 
     return {};
   }
@@ -234,8 +254,6 @@ public:
     std::unique_lock Lock(FdMutex);
     if (auto It = FdMap.find(Fd); It == FdMap.end()) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else if (It->second->isPreopened()) {
-      return WasiUnexpect(__WASI_ERRNO_NOTSUP);
     } else {
       FdMap.erase(It);
       return {};
@@ -507,12 +525,8 @@ public:
     std::unique_lock Lock(FdMutex);
     if (auto It = FdMap.find(Fd); It == FdMap.end()) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else if (It->second->isPreopened()) {
-      return WasiUnexpect(__WASI_ERRNO_NOTSUP);
     } else if (auto It2 = FdMap.find(To); It2 == FdMap.end()) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else if (It2->second->isPreopened()) {
-      return WasiUnexpect(__WASI_ERRNO_NOTSUP);
     } else {
       FdMap.erase(It2);
       auto Node = FdMap.extract(It);
@@ -600,8 +614,11 @@ public:
   /// @param[in] Path The path at which to create the directory.
   /// @return Nothing or WASI error
   WasiExpect<void> pathCreateDirectory(__wasi_fd_t Fd, std::string_view Path) {
+    if (!VINode::isPathValid(Path)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto Node = getNodeOrNull(Fd);
-    return VINode::pathCreateDirectory(FS, std::move(Node), Path);
+    return VINode::pathCreateDirectory(std::move(Node), Path);
   }
 
   /// Return the attributes of a file or directory.
@@ -617,8 +634,11 @@ public:
   WasiExpect<void> pathFilestatGet(__wasi_fd_t Fd, std::string_view Path,
                                    __wasi_lookupflags_t Flags,
                                    __wasi_filestat_t &Filestat) {
+    if (!VINode::isPathValid(Path)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto Node = getNodeOrNull(Fd);
-    return VINode::pathFilestatGet(FS, std::move(Node), Path, Flags, Filestat);
+    return VINode::pathFilestatGet(std::move(Node), Path, Flags, Filestat);
   }
 
   /// Adjust the timestamps of a file or directory.
@@ -638,8 +658,11 @@ public:
                                         __wasi_timestamp_t ATim,
                                         __wasi_timestamp_t MTim,
                                         __wasi_fstflags_t FstFlags) {
+    if (!VINode::isPathValid(Path)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto Node = getNodeOrNull(Fd);
-    return VINode::pathFilestatSetTimes(FS, std::move(Node), Path, Flags, ATim,
+    return VINode::pathFilestatSetTimes(std::move(Node), Path, Flags, ATim,
                                         MTim, FstFlags);
   }
 
@@ -659,9 +682,15 @@ public:
   WasiExpect<void> pathLink(__wasi_fd_t Old, std::string_view OldPath,
                             __wasi_fd_t New, std::string_view NewPath,
                             __wasi_lookupflags_t LookupFlags) {
+    if (!VINode::isPathValid(OldPath)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
+    if (!VINode::isPathValid(NewPath)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto OldNode = getNodeOrNull(Old);
     auto NewNode = getNodeOrNull(New);
-    return VINode::pathLink(FS, std::move(OldNode), OldPath, std::move(NewNode),
+    return VINode::pathLink(std::move(OldNode), OldPath, std::move(NewNode),
                             NewPath, LookupFlags);
   }
 
@@ -701,9 +730,12 @@ public:
                                    __wasi_rights_t FsRightsBase,
                                    __wasi_rights_t FsRightsInheriting,
                                    __wasi_fdflags_t FdFlags) {
+    if (!VINode::isPathValid(Path)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto Node = getNodeOrNull(Fd);
     if (auto Res =
-            VINode::pathOpen(FS, std::move(Node), Path, LookupFlags, OpenFlags,
+            VINode::pathOpen(std::move(Node), Path, LookupFlags, OpenFlags,
                              FsRightsBase, FsRightsInheriting, FdFlags);
         unlikely(!Res)) {
       return WasiUnexpect(Res);
@@ -727,8 +759,11 @@ public:
   /// @return Nothing or WASI error.
   WasiExpect<void> pathReadlink(__wasi_fd_t Fd, std::string_view Path,
                                 Span<char> Buffer, __wasi_size_t &NRead) {
+    if (!VINode::isPathValid(Path)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto Node = getNodeOrNull(Fd);
-    return VINode::pathReadlink(FS, std::move(Node), Path, Buffer, NRead);
+    return VINode::pathReadlink(std::move(Node), Path, Buffer, NRead);
   }
 
   /// Remove a directory.
@@ -742,8 +777,11 @@ public:
   /// @param[in] Path The path to a directory to remove.
   /// @return Nothing or WASI error.
   WasiExpect<void> pathRemoveDirectory(__wasi_fd_t Fd, std::string_view Path) {
+    if (!VINode::isPathValid(Path)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto Node = getNodeOrNull(Fd);
-    return VINode::pathRemoveDirectory(FS, std::move(Node), Path);
+    return VINode::pathRemoveDirectory(std::move(Node), Path);
   }
 
   /// Rename a file or directory.
@@ -760,10 +798,16 @@ public:
   /// @return Nothing or WASI error.
   WasiExpect<void> pathRename(__wasi_fd_t Old, std::string_view OldPath,
                               __wasi_fd_t New, std::string_view NewPath) {
+    if (!VINode::isPathValid(OldPath)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
+    if (!VINode::isPathValid(NewPath)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto OldNode = getNodeOrNull(Old);
     auto NewNode = getNodeOrNull(New);
-    return VINode::pathRename(FS, std::move(OldNode), OldPath,
-                              std::move(NewNode), NewPath);
+    return VINode::pathRename(std::move(OldNode), OldPath, std::move(NewNode),
+                              NewPath);
   }
 
   /// Create a symbolic link.
@@ -778,8 +822,14 @@ public:
   /// @return Nothing or WASI error
   WasiExpect<void> pathSymlink(std::string_view OldPath, __wasi_fd_t New,
                                std::string_view NewPath) {
+    if (!VINode::isPathValid(OldPath)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
+    if (!VINode::isPathValid(NewPath)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto NewNode = getNodeOrNull(New);
-    return VINode::pathSymlink(FS, OldPath, std::move(NewNode), NewPath);
+    return VINode::pathSymlink(OldPath, std::move(NewNode), NewPath);
   }
 
   /// Unlink a file.
@@ -793,22 +843,24 @@ public:
   /// @param[in] Path The path to a file to unlink.
   /// @return Nothing or WASI error.
   WasiExpect<void> pathUnlinkFile(__wasi_fd_t Fd, std::string_view Path) {
+    if (!VINode::isPathValid(Path)) {
+      return WasiUnexpect(__WASI_ERRNO_INVAL);
+    }
     auto Node = getNodeOrNull(Fd);
-    return VINode::pathUnlinkFile(FS, std::move(Node), Path);
+    return VINode::pathUnlinkFile(std::move(Node), Path);
   }
 
-  /// Concurrently poll for the occurrence of a set of events.
+  /// Acquire a Poller for concurrently poll for the occurrence of a set of
+  /// events.
   ///
-  /// @param[in] NSubscriptions Both the number of subscriptions and events.
+  /// @param[in] Events The output buffer for events.
   /// @return Poll helper or WASI error.
-  WasiExpect<EVPoller> pollOneoff(__wasi_size_t NSubscriptions) noexcept;
+  WasiExpect<EVPoller> acquirePoller(Span<__wasi_event_t> Events) noexcept;
 
-  /// Concurrently poll for the occurrence of a set of events in edge-triggered
-  /// mode.
+  /// Release a used Poller object.
   ///
-  /// @param[in] NSubscriptions Both the number of subscriptions and events.
-  /// @return Poll helper or WASI error.
-  WasiExpect<EVEpoller> epollOneoff(__wasi_size_t NSubscriptions) noexcept;
+  /// @param[in] Poller Used poller object.
+  void releasePoller(EVPoller &&Poller) noexcept;
 
   /// Terminate the process normally. An exit code of 0 indicates successful
   /// termination of the program. The meanings of other values is dependent on
@@ -864,8 +916,7 @@ public:
                                    __wasi_sock_type_t SockType) noexcept {
 
     std::shared_ptr<VINode> Node;
-    if (auto Res = VINode::sockOpen(FS, AddressFamily, SockType);
-        unlikely(!Res)) {
+    if (auto Res = VINode::sockOpen(AddressFamily, SockType); unlikely(!Res)) {
       return WasiUnexpect(Res);
     } else {
       Node = std::move(*Res);
@@ -874,23 +925,15 @@ public:
     return generateRandomFdToNode(Node);
   }
 
-  WasiExpect<void> sockBindV1(__wasi_fd_t Fd, uint8_t *Address,
-                              uint8_t AddressLength, uint16_t Port) noexcept {
+  WasiExpect<void> sockBind(__wasi_fd_t Fd,
+                            __wasi_address_family_t AddressFamily,
+                            Span<const uint8_t> Address,
+                            uint16_t Port) noexcept {
     auto Node = getNodeOrNull(Fd);
     if (unlikely(!Node)) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
     } else {
-      return Node->sockBindV1(Address, AddressLength, Port);
-    }
-  }
-
-  WasiExpect<void> sockBindV2(__wasi_fd_t Fd, uint8_t *Address,
-                              uint8_t AddressLength, uint16_t Port) noexcept {
-    auto Node = getNodeOrNull(Fd);
-    if (unlikely(!Node)) {
-      return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else {
-      return Node->sockBindV2(Address, AddressLength, Port);
+      return Node->sockBind(AddressFamily, Address, Port);
     }
   }
 
@@ -903,11 +946,12 @@ public:
     }
   }
 
-  WasiExpect<__wasi_fd_t> sockAcceptV1(__wasi_fd_t Fd) noexcept {
+  WasiExpect<__wasi_fd_t> sockAccept(__wasi_fd_t Fd,
+                                     __wasi_fdflags_t FdFlags) noexcept {
     auto Node = getNodeOrNull(Fd);
     std::shared_ptr<VINode> NewNode;
 
-    if (auto Res = Node->sockAcceptV1(); unlikely(!Res)) {
+    if (auto Res = Node->sockAccept(FdFlags); unlikely(!Res)) {
       return WasiUnexpect(Res);
     } else {
       NewNode = std::move(*Res);
@@ -916,39 +960,15 @@ public:
     return generateRandomFdToNode(NewNode);
   }
 
-  WasiExpect<__wasi_fd_t> sockAcceptV2(__wasi_fd_t Fd,
-                                       __wasi_fdflags_t FdFlags) noexcept {
-    auto Node = getNodeOrNull(Fd);
-    std::shared_ptr<VINode> NewNode;
-
-    if (auto Res = Node->sockAcceptV2(FdFlags); unlikely(!Res)) {
-      return WasiUnexpect(Res);
-    } else {
-      NewNode = std::move(*Res);
-    }
-
-    return generateRandomFdToNode(NewNode);
-  }
-
-  WasiExpect<void> sockConnectV1(__wasi_fd_t Fd, uint8_t *Address,
-                                 uint8_t AddressLength,
-                                 uint16_t Port) noexcept {
+  WasiExpect<void> sockConnect(__wasi_fd_t Fd,
+                               __wasi_address_family_t AddressFamily,
+                               Span<const uint8_t> Address,
+                               uint16_t Port) noexcept {
     auto Node = getNodeOrNull(Fd);
     if (unlikely(!Node)) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
     } else {
-      return Node->sockConnectV1(Address, AddressLength, Port);
-    }
-  }
-
-  WasiExpect<void> sockConnectV2(__wasi_fd_t Fd, uint8_t *Address,
-                                 uint8_t AddressLength,
-                                 uint16_t Port) noexcept {
-    auto Node = getNodeOrNull(Fd);
-    if (unlikely(!Node)) {
-      return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else {
-      return Node->sockConnectV2(Address, AddressLength, Port);
+      return Node->sockConnect(AddressFamily, Address, Port);
     }
   }
 
@@ -974,58 +994,30 @@ public:
   }
 
   /// Receive a message from a socket.
-  /// This is a V1 version, which doesn't return the Port in the very first
-  /// version.
   ///
-  /// Note: This is similar to `recv` in POSIX, though it also supports reading
-  /// the data into multiple buffers in the manner of `readv`.
+  /// Note: This is similar to `recvfrom` in POSIX, though it also supports
+  /// reading the data into multiple buffers in the manner of `readv`.
   ///
   /// @param[in] RiData List of scatter/gather vectors to which to store data.
   /// @param[in] RiFlags Message flags.
-  /// @param[in] Address Address of the target.
-  /// @param[in] AddressLength The buffer size of Address.
+  /// @param[out] AddressFamilyPtr The pointer to store address family.
+  /// @param[out] Address The buffer to store address.
+  /// @param[out] PortPtr The pointer to store port.
   /// @param[out] NRead Return the number of bytes stored in RiData.
   /// @param[out] RoFlags Return message flags.
   /// @return Nothing or WASI error.
-  WasiExpect<void> sockRecvFromV1(__wasi_fd_t Fd, Span<Span<uint8_t>> RiData,
-                                  __wasi_riflags_t RiFlags, uint8_t *Address,
-                                  uint8_t AddressLength, __wasi_size_t &NRead,
-                                  __wasi_roflags_t &RoFlags) const noexcept {
+  WasiExpect<void> sockRecvFrom(__wasi_fd_t Fd, Span<Span<uint8_t>> RiData,
+                                __wasi_riflags_t RiFlags,
+                                __wasi_address_family_t *AddressFamilyPtr,
+                                Span<uint8_t> Address, uint16_t *PortPtr,
+                                __wasi_size_t &NRead,
+                                __wasi_roflags_t &RoFlags) const noexcept {
     auto Node = getNodeOrNull(Fd);
     if (unlikely(!Node)) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
     } else {
-      return Node->sockRecvFromV1(RiData, RiFlags, Address, AddressLength,
-                                  NRead, RoFlags);
-    }
-  }
-
-  /// Receive a message from a socket.
-  /// This is the V2 version, which is almost the same as V1, except for
-  /// returning the Port information.
-  ///
-  /// Note: This is similar to `recv` in POSIX, though it also supports reading
-  /// the data into multiple buffers in the manner of `readv`.
-  ///
-  /// @param[in] RiData List of scatter/gather vectors to which to store data.
-  /// @param[in] RiFlags Message flags.
-  /// @param[in] Address Address of the target.
-  /// @param[in] AddressLength The buffer size of Address.
-  /// @param[out] PortPtr The address to store port.
-  /// @param[out] NRead Return the number of bytes stored in RiData.
-  /// @param[out] RoFlags Return message flags.
-  /// @return Nothing or WASI error.
-  WasiExpect<void> sockRecvFromV2(__wasi_fd_t Fd, Span<Span<uint8_t>> RiData,
-                                  __wasi_riflags_t RiFlags, uint8_t *Address,
-                                  uint8_t AddressLength, uint32_t *PortPtr,
-                                  __wasi_size_t &NRead,
-                                  __wasi_roflags_t &RoFlags) const noexcept {
-    auto Node = getNodeOrNull(Fd);
-    if (unlikely(!Node)) {
-      return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else {
-      return Node->sockRecvFromV2(RiData, RiFlags, Address, AddressLength,
-                                  PortPtr, NRead, RoFlags);
+      return Node->sockRecvFrom(RiData, RiFlags, AddressFamilyPtr, Address,
+                                PortPtr, NRead, RoFlags);
     }
   }
 
@@ -1058,19 +1050,21 @@ public:
   /// @param[in] SiData List of scatter/gather vectors to which to retrieve
   /// data.
   /// @param[in] SiFlags Message flags.
+  /// @param[in] AddressFamily Address family of the target.
   /// @param[in] Address Address of the target.
-  /// @param[in] AddressLength The buffer size of Address.
+  /// @param[in] Port Connected port.
   /// @param[out] NWritten The number of bytes transmitted.
   /// @return Nothing or WASI error
   WasiExpect<void> sockSendTo(__wasi_fd_t Fd, Span<Span<const uint8_t>> SiData,
-                              __wasi_siflags_t SiFlags, uint8_t *Address,
-                              uint8_t AddressLength, int32_t Port,
+                              __wasi_siflags_t SiFlags,
+                              __wasi_address_family_t AddressFamily,
+                              Span<const uint8_t> Address, uint16_t Port,
                               __wasi_size_t &NWritten) const noexcept {
     auto Node = getNodeOrNull(Fd);
     if (unlikely(!Node)) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
     } else {
-      return Node->sockSendTo(SiData, SiFlags, Address, AddressLength, Port,
+      return Node->sockSendTo(SiData, SiFlags, AddressFamily, Address, Port,
                               NWritten);
     }
   }
@@ -1093,79 +1087,51 @@ public:
 
   WasiExpect<void> sockGetOpt(__wasi_fd_t Fd,
                               __wasi_sock_opt_level_t SockOptLevel,
-                              __wasi_sock_opt_so_t SockOptName, void *FlagPtr,
-                              uint32_t *FlagSizePtr) const noexcept {
+                              __wasi_sock_opt_so_t SockOptName,
+                              Span<uint8_t> &Flag) const noexcept {
     auto Node = getNodeOrNull(Fd);
     if (unlikely(!Node)) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
     } else {
-      return Node->sockGetOpt(SockOptLevel, SockOptName, FlagPtr, FlagSizePtr);
+      return Node->sockGetOpt(SockOptLevel, SockOptName, Flag);
     }
   }
 
   WasiExpect<void> sockSetOpt(__wasi_fd_t Fd,
                               __wasi_sock_opt_level_t SockOptLevel,
-                              __wasi_sock_opt_so_t SockOptName, void *FlagPtr,
-                              uint32_t FlagSizePtr) const noexcept {
+                              __wasi_sock_opt_so_t SockOptName,
+                              Span<const uint8_t> Flag) const noexcept {
     auto Node = getNodeOrNull(Fd);
     if (unlikely(!Node)) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
     } else {
-      return Node->sockSetOpt(SockOptLevel, SockOptName, FlagPtr, FlagSizePtr);
+      return Node->sockSetOpt(SockOptLevel, SockOptName, Flag);
     }
   }
 
   /// Return the address and port of the file descriptor.
-  /// The V1 will also return the type of the address, so users can
-  /// identify the address type from `AddrTypePtr`.
-  WasiExpect<void> sockGetLocalAddrV1(__wasi_fd_t Fd, uint8_t *Address,
-                                      uint32_t *AddrTypePtr,
-                                      uint32_t *PortPtr) const noexcept {
+  WasiExpect<void> sockGetLocalAddr(__wasi_fd_t Fd,
+                                    __wasi_address_family_t *AddressFamilyPtr,
+                                    Span<uint8_t> Address,
+                                    uint16_t *PortPtr) const noexcept {
     auto Node = getNodeOrNull(Fd);
     if (unlikely(!Node)) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
     } else {
-      return Node->sockGetLocalAddrV1(Address, AddrTypePtr, PortPtr);
-    }
-  }
-
-  /// Return the address and port of the file descriptor.
-  /// The the new address type will become an unified size, there is
-  /// no need to return the address type.
-  WasiExpect<void> sockGetLocalAddrV2(__wasi_fd_t Fd, uint8_t *Address,
-                                      uint32_t *PortPtr) const noexcept {
-    auto Node = getNodeOrNull(Fd);
-    if (unlikely(!Node)) {
-      return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else {
-      return Node->sockGetLocalAddrV2(Address, PortPtr);
+      return Node->sockGetLocalAddr(AddressFamilyPtr, Address, PortPtr);
     }
   }
 
   /// Retrieve the remote address and port from the given file descriptor.
-  /// The V1 will also return the type of the address, so users can
-  /// identify the address type from `AddrTypePtr`.
-  WasiExpect<void> sockGetPeerAddrV1(__wasi_fd_t Fd, uint8_t *Address,
-                                     uint32_t *AddrTypePtr,
-                                     uint32_t *PortPtr) const noexcept {
+  WasiExpect<void> sockGetPeerAddr(__wasi_fd_t Fd,
+                                   __wasi_address_family_t *AddressFamilyPtr,
+                                   Span<uint8_t> Address,
+                                   uint16_t *PortPtr) const noexcept {
     auto Node = getNodeOrNull(Fd);
     if (unlikely(!Node)) {
       return WasiUnexpect(__WASI_ERRNO_BADF);
     } else {
-      return Node->sockGetPeerAddrV1(Address, AddrTypePtr, PortPtr);
-    }
-  }
-
-  /// Retrieve the remote address and port from the given file descriptor.
-  /// The the new address type will become an unified size, there is
-  /// no need to return the address type.
-  WasiExpect<void> sockGetPeerAddrV2(__wasi_fd_t Fd, uint8_t *Address,
-                                     uint32_t *PortPtr) const noexcept {
-    auto Node = getNodeOrNull(Fd);
-    if (unlikely(!Node)) {
-      return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else {
-      return Node->sockGetPeerAddrV2(Address, PortPtr);
+      return Node->sockGetPeerAddr(AddressFamilyPtr, Address, PortPtr);
     }
   }
 
@@ -1197,19 +1163,15 @@ public:
 private:
   std::vector<std::string> Arguments;
   std::vector<std::string> EnvironVariables;
-  VFS FS;
   __wasi_exitcode_t ExitCode = 0;
 
+  mutable std::shared_mutex PollerMutex; ///< Protect PollerPool
+  std::vector<EVPoller> PollerPool;
+  friend class EVPoller;
+
   mutable std::shared_mutex FdMutex; ///< Protect FdMap
   std::unordered_map<__wasi_fd_t, std::shared_ptr<VINode>> FdMap;
 
-  // unique epoll fd;
-  int RegistrationFd = -1;
-  std::unordered_map<int, uint32_t> Registration;
-
-  friend class EVPoller;
-  friend class EVEpoller;
-
   std::shared_ptr<VINode> getNodeOrNull(__wasi_fd_t Fd) const {
     std::shared_lock Lock(FdMutex);
     if (auto It = FdMap.find(Fd); It != FdMap.end()) {
@@ -1233,99 +1195,79 @@ private:
   }
 };
 
-class EVPoller : private VPoller {
+class EVPoller : protected VPoller {
 public:
+  EVPoller(EVPoller &&) = default;
+  EVPoller &operator=(EVPoller &&) = default;
+
   using VPoller::clock;
+  using VPoller::error;
+  using VPoller::prepare;
+  using VPoller::reset;
+  using VPoller::result;
+  using VPoller::VPoller;
   using VPoller::wait;
 
-  EVPoller(VPoller &&P, Environ &E) : VPoller(std::move(P)), Env(E) {}
-
-  WasiExpect<void> clock(__wasi_clockid_t Clock, __wasi_timestamp_t Timeout,
-                         __wasi_timestamp_t Precision,
-                         __wasi_subclockflags_t Flags,
-                         __wasi_userdata_t UserData) noexcept {
-    return VPoller::clock(Clock, Timeout, Precision, Flags, UserData);
-  }
-
-  WasiExpect<void> read(__wasi_fd_t Fd, __wasi_userdata_t UserData) noexcept {
-    auto Node = Env.getNodeOrNull(Fd);
-    if (unlikely(!Node)) {
-      return WasiUnexpect(__WASI_ERRNO_BADF);
+  /// Concurrently poll for a ready-to-read event.
+  ///
+  /// @param[in] Fd The file descriptor on which to wait for it to become ready
+  /// for reading.
+  /// @param[in] Trigger Specifying whether the notification is level-trigger or
+  /// edge-trigger.
+  /// @param[in] UserData User-provided value that may be attached to objects
+  /// that is retained when extracted from the implementation.
+  void read(__wasi_fd_t Fd, TriggerType Trigger,
+            __wasi_userdata_t UserData) noexcept {
+    if (auto Node = env().getNodeOrNull(Fd); unlikely(!Node)) {
+      VPoller::error(UserData, __WASI_ERRNO_BADF, __WASI_EVENTTYPE_FD_READ);
     } else {
-      return VPoller::read(Node, UserData);
+      VPoller::read(Node, Trigger, UserData);
     }
   }
 
-  WasiExpect<void> write(__wasi_fd_t Fd, __wasi_userdata_t UserData) noexcept {
-    auto Node = Env.getNodeOrNull(Fd);
-    if (unlikely(!Node)) {
-      return WasiUnexpect(__WASI_ERRNO_BADF);
+  /// Concurrently poll for a ready-to-write event.
+  ///
+  /// @param[in] Fd The file descriptor on which to wait for it to become ready
+  /// for writing.
+  /// @param[in] Trigger Specifying whether the notification is level-trigger or
+  /// edge-trigger.
+  /// @param[in] UserData User-provided value that may be attached to objects
+  /// that is retained when extracted from the implementation.
+  void write(__wasi_fd_t Fd, TriggerType Trigger,
+             __wasi_userdata_t UserData) noexcept {
+    if (auto Node = env().getNodeOrNull(Fd); unlikely(!Node)) {
+      VPoller::error(UserData, __WASI_ERRNO_BADF, __WASI_EVENTTYPE_FD_WRITE);
     } else {
-      return VPoller::write(Node, UserData);
+      VPoller::write(Node, Trigger, UserData);
     }
   }
 
 private:
-  Environ &Env;
-};
-
-class EVEpoller : private VEpoller {
-public:
-  using VEpoller::CallbackType;
-  using VEpoller::clock;
-  using VEpoller::getFd;
-  using VEpoller::wait;
-
-  EVEpoller(VEpoller &&P, Environ &E) : VEpoller(std::move(P)), Env(E) {}
-
-  WasiExpect<void> clock(__wasi_clockid_t Clock, __wasi_timestamp_t Timeout,
-                         __wasi_timestamp_t Precision,
-                         __wasi_subclockflags_t Flags,
-                         __wasi_userdata_t UserData) noexcept {
-    return VEpoller::clock(Clock, Timeout, Precision, Flags, UserData);
-  }
-
-  WasiExpect<void> read(__wasi_fd_t Fd, __wasi_userdata_t UserData) noexcept {
-    auto Node = Env.getNodeOrNull(Fd);
-    if (unlikely(!Node)) {
-      return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else {
-      return VEpoller::read(Node, UserData, Env.Registration);
-    }
-  }
-
-  WasiExpect<void> write(__wasi_fd_t Fd, __wasi_userdata_t UserData) noexcept {
-    auto Node = Env.getNodeOrNull(Fd);
-    if (unlikely(!Node)) {
-      return WasiUnexpect(__WASI_ERRNO_BADF);
-    } else {
-      return VEpoller::write(Node, UserData, Env.Registration);
-    }
-  }
-  WasiExpect<void> wait(CallbackType Callback) noexcept {
-    return VEpoller::wait(Callback, Env.Registration);
-  }
-
-  int getFd() noexcept { return VEpoller::getFd(); }
-
-private:
-  Environ &Env;
+  Environ &env() noexcept { return static_cast<Environ &>(Ctx.get()); }
 };
 
 inline WasiExpect<EVPoller>
-Environ::pollOneoff(__wasi_size_t NSubscriptions) noexcept {
-  return VINode::pollOneoff(NSubscriptions).map([this](VPoller &&P) {
-    return EVPoller(std::move(P), *this);
-  });
+Environ::acquirePoller(Span<__wasi_event_t> Events) noexcept {
+  auto Poller = [this]() noexcept -> EVPoller {
+    std::unique_lock Lock(PollerMutex);
+    if (PollerPool.empty()) {
+      return EVPoller(*this);
+    } else {
+      EVPoller Result(std::move(PollerPool.back()));
+      PollerPool.pop_back();
+      return Result;
+    }
+  }();
+
+  if (auto Res = Poller.prepare(Events); !Res) {
+    return WasiUnexpect(Res);
+  }
+  return Poller;
 }
 
-inline WasiExpect<EVEpoller>
-Environ::epollOneoff(__wasi_size_t NSubscriptions) noexcept {
-  auto Evepoller =
-      VINode::epollOneoff(NSubscriptions, RegistrationFd)
-          .map([this](VEpoller &&P) { return EVEpoller(std::move(P), *this); });
-  RegistrationFd = Evepoller.value().getFd();
-  return Evepoller;
+inline void Environ::releasePoller(EVPoller &&Poller) noexcept {
+  std::unique_lock Lock(PollerMutex);
+  PollerPool.push_back(std::move(Poller));
 }
 
 } // namespace WASI
