@@ -7,12 +7,11 @@
 #include "common/version.h"
 #include "po/argument_parser.h"
 
+#include <algorithm>
 #include <cstdio>
 #include <iostream>
 #include <vector>
 
-#include <boost/algorithm/searching/boyer_moore_horspool.hpp>
-
 namespace WasmEdge {
 namespace Driver {
 
@@ -122,13 +121,13 @@ int FuzzPO(const uint8_t *Data, size_t Size) noexcept {
 
   static constexpr const std::array<char, 4> Separator = {'\xde', '\xad',
                                                           '\xbe', '\xef'};
-  static const boost::algorithm::boyer_moore_horspool Searcher(
-      Separator.begin(), Separator.end());
+  static const std::boyer_moore_horspool_searcher Searcher(Separator.begin(),
+                                                           Separator.end());
   Span<const char> RawArgs(reinterpret_cast<const char *>(Data), Size);
   std::vector<std::string> ArgvStr;
   std::vector<const char *> Argv;
   while (!RawArgs.empty()) {
-    const auto It = Searcher(RawArgs.begin(), RawArgs.end()).first;
+    const auto It = std::search(RawArgs.begin(), RawArgs.end(), Searcher);
     ArgvStr.emplace_back(RawArgs.begin(), It);
     RawArgs = RawArgs.subspan(std::min<size_t>(
         std::distance(RawArgs.begin(), It) + 4, RawArgs.size()));
